#!/usr/bin/python
import commands
import ujson
import os
import shutil
import sys
import time
import glob

import debinterface.interfaces
import edge.admin.ui_types as ui_types
from edge import eventlog
import jsonschema
from edge.admin.reply_helpers import *
from edge.com.fs import ln_schema, try_lock_pid_file
from edge.com import net

UI_NETWORK_SCHEMA_FILEPATH = '/etc/edge/setupstore/schemas/network.json'
UI_NETWORK_JNODE_FILEPATH = '/etc/edge/setupstore/admin/actions/network.json'
UI_NETWORK_DEFJNODE_FILEPATH = '/opt/edge/share/setupstore/ui/network-jnode.json'

MANAGEMENT_MAC_FILEPATH = '/etc/network/management-mac'

LOCK_FILEPATH = '/var/run/edge/network'

INTERFACES_PATH = '/etc/network/interfaces'
ETC_HOSTNAME_FILEPATH = '/etc/hostname'
ETC_HOSTS_FILEPATH = '/etc/hosts'

RESTART_NONE = 0
RESTART_HOSTNAME = 1
RESTART_NETWORKING = 2

BACKUP_DIRPATH = '/etc/edge/setupstore/backup/network.' + str(round(time.time() * 1000000)) + '/'


def try_create_backup():
    try:
        os.makedirs(BACKUP_DIRPATH)
        shutil.copy(ETC_HOSTNAME_FILEPATH, BACKUP_DIRPATH)
        shutil.copy(ETC_HOSTS_FILEPATH, BACKUP_DIRPATH)
        shutil.copy(UI_NETWORK_JNODE_FILEPATH, BACKUP_DIRPATH)
        return True

    except Exception:

        return False


def try_restore_backup():
    try:
        shutil.copy(BACKUP_DIRPATH + 'hostname', ETC_HOSTNAME_FILEPATH)
        shutil.copy(BACKUP_DIRPATH + 'hosts', ETC_HOSTS_FILEPATH)
        shutil.copy(BACKUP_DIRPATH + 'network.json', UI_NETWORK_JNODE_FILEPATH)
        os.rmdir(BACKUP_DIRPATH)
        return True

    except Exception:

        return False


def remove_backup():
    try:
        shutil.rmtree(BACKUP_DIRPATH)
    except Exception:
        pass


def edit_hostname(old_hostname, new_hostname):

    if old_hostname == new_hostname:
        return RESTART_NONE

    with open(ETC_HOSTS_FILEPATH, 'r') as old_hosts_file:
        old_hosts_lines = old_hosts_file.readlines()

    line_found = False
    new_host_lines = []
    for old_line in old_hosts_lines:
        stripped_old_line = old_line.strip()
        if not stripped_old_line or stripped_old_line.startswith('#'):
            new_host_lines.append(stripped_old_line)
            continue

        param_set = stripped_old_line.split()
        if len(param_set) < 2:
            new_host_lines.append(stripped_old_line)
            continue

        hostname = param_set[1]
        if hostname == old_hostname:
            param_set[1] = new_hostname
            new_host_lines.append('\t'.join(param_set))
            line_found = True

        else:
            new_host_lines.append(stripped_old_line)

    if not line_found:
        new_host_lines.append('\t'.join(['127.0.0.1', new_hostname]))

    with open(ETC_HOSTS_FILEPATH, 'w') as new_hosts_file:
        new_hosts_file.write('\n'.join(new_host_lines))

    with open(ETC_HOSTNAME_FILEPATH, 'w') as hostname_file:
        hostname_file.write(new_hostname)

    return RESTART_HOSTNAME


def check_iface_changed(old_jnode, new_jnode):
    if old_jnode['ip-mode'] != new_jnode['ip-mode']:
        return True

    if old_jnode['ip-mode'] == 'static':
        if old_jnode['ip-address'] != new_jnode['ip-address']:
            return True

        if old_jnode['ip-mask'] != new_jnode['ip-mask']:
            return True

        if old_jnode['ip-gateway'] != new_jnode['ip-gateway']:
            return True
    else:
        pass

    if old_jnode['dns-mode'] != new_jnode['dns-mode']:
        return True

    if old_jnode['dns-mode'] == 'static':
        if old_jnode['dns-1'] != new_jnode['dns-1']:
            return True

        if old_jnode['dns-2'] != new_jnode['dns-2']:
            return True
    else:
        pass

    return False


def edit_iface(old_jnode, new_jnode):
    if not check_iface_changed(old_jnode, new_jnode):
        return RESTART_NONE

    mac_address = new_jnode['mac']

    ifname = net.get_ifname(mac_address)

    iface_conf = debinterface.interfaces.Interfaces(interfaces_path=INTERFACES_PATH)

    adapter_list = filter(lambda item: item.export(options_list=['name'])['name'] == ifname, iface_conf.adapters)

    if len(adapter_list) != 1:
        raise ValueError
    adapter = adapter_list[0]

    adapter.reset()
    if new_jnode['ip-mode'] == 'static':
        adapter.setName(ifname)
        adapter.setAuto(True)
        adapter.setAddrFam('inet')
        adapter.setAddressSource(new_jnode['ip-mode'])
        adapter.setAddress(new_jnode['ip-address'])
        adapter.setNetmask(new_jnode['ip-mask'])
        adapter.setGateway(new_jnode['ip-gateway'])

    elif new_jnode['ip-mode'] == 'dhcp':
        adapter.setName(ifname)
        adapter.setAuto(True)
        adapter.setAddrFam('inet')
        adapter.setAddressSource(new_jnode['ip-mode'])

    else:
        raise NotImplemented('unsupported ip-mode %s' % new_jnode['ip-mode'])

    if new_jnode['dns-mode'] == 'static':
        dns_1 = new_jnode['dns-1']
        if dns_1:
            adapter.setDnsNameservers(dns_1)

        dns_2 = new_jnode['dns-2']
        if dns_2:
            adapter.setDnsNameservers(dns_1 + ' ' + dns_2)

    elif new_jnode['dns-mode'] == 'dhcp':
        pass

    else:
        raise NotImplemented('unsupported dns-mode %s' % new_jnode['dns-mode'])

    iface_conf.writeInterfaces()

    return RESTART_NETWORKING


def _handle_get(query_args):
    """
    :type query_args: UiQueryArgs
    :rtype: dict
    """
    if query_args.path_info:
        eventlog.log_d('fail: unexpected path-info')
        return reply_fail(REPLY_NOT_FOUND_404)

    if query_args.query_string:
        eventlog.log_d('fail: unexpected query-string')
        return reply_fail(REPLY_BAD_REQUEST_400)

    # create ui-jnode
    if not os.path.exists(UI_NETWORK_JNODE_FILEPATH):
        # noinspection PyBroadException
        try:
            with open(MANAGEMENT_MAC_FILEPATH, 'r') as mgmt_mac_file:
                mac_addr = mgmt_mac_file.readlines()[0].strip()

        except IOError:
            eventlog.log_d('management-mac file not found, using eth0 as management interface')
            mac_addr = net.get_ifmac('eth0')
            if not mac_addr:
                eventlog.log_d('fail: get ifmac')
                return reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)

        # noinspection PyBroadException
        try:
            hostname = commands.getoutput("hostname")

            with open(UI_NETWORK_DEFJNODE_FILEPATH, 'r') as json_file:
                def_ui_jnode = ujson.load(json_file)
            eventlog.log_d(def_ui_jnode)
            def_ui_jnode['ro']['mac'] = mac_addr
            def_ui_jnode['rw']['hostname'] = hostname

            with open(UI_NETWORK_JNODE_FILEPATH, 'w') as json_file:
                ujson.dump(def_ui_jnode, json_file, indent=2)

        except:
            eventlog.log_x()
            return reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)

    # create ui-schema
    if not os.path.islink(UI_NETWORK_SCHEMA_FILEPATH) or not os.path.exists(os.readlink(UI_NETWORK_SCHEMA_FILEPATH)):
        if not os.path.exists(UI_NETWORK_JNODE_FILEPATH):
            eventlog.log_d('fail: %s not found' % UI_NETWORK_JNODE_FILEPATH)
            return reply_fail(REPLY_NOT_FOUND_404)

        ln_schema(UI_NETWORK_SCHEMA_FILEPATH, UI_NETWORK_JNODE_FILEPATH)

    try:
        with open(UI_NETWORK_JNODE_FILEPATH, 'r') as json_file:
            ui_jnode = ujson.load(json_file)

        ui_jnode['rw']['mac'] = ui_jnode['ro']['mac']
        return reply_done(ui_jnode['rw'])

    except (IOError, ValueError) as x:
        eventlog.log_d('fail: load ui-jnode (%s)' % x.message)
        return reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)


def _handle_put(raw_ui_jnode, new_ui_jnode):
    """
    :type raw_ui_jnode: dict
    :typeype new_ui_jnode:  dict
    :rtype: dict
    """
    # load ui-jnode-schema
    try:
        with open(UI_NETWORK_SCHEMA_FILEPATH, 'r') as json_file:
            ui_schema_jnode = ujson.load(json_file)

    except (IOError, ValueError) as x:
        eventlog.log_d("fail: load ui-jnode-schema (%s)" % x.message)
        return reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)

    # validate new-ui-jnode
    try:
        jsonschema.validate(new_ui_jnode, ui_schema_jnode)

    except (ValueError, jsonschema.ValidationError) as x:
        eventlog.log_d("fail: validate new-ui-jnode (%s)" % x.message)
        return reply_fail(REPLY_BAD_REQUEST_400)

    if not try_lock_pid_file(LOCK_FILEPATH):
        return reply_fail(REPLY_CONFLICT_409)

    if not try_create_backup():
        return reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)

    old_ui_jnode = raw_ui_jnode['rw']

    success = False
    try:
        restart_mask = 0
        restart_mask |= edit_iface(old_ui_jnode, new_ui_jnode) or restart_mask
        restart_mask |= edit_hostname(old_ui_jnode['hostname'], new_ui_jnode['hostname']) or restart_mask

        raw_ui_jnode['rw'] = new_ui_jnode.copy()
        raw_ui_jnode['rw'].pop('mac')
        with open(UI_NETWORK_JNODE_FILEPATH, 'w') as f:
            f.write((ujson.dumps(raw_ui_jnode, indent=2)))

        if restart_mask & RESTART_HOSTNAME:
            os.system("/etc/init.d/hostname.sh start")

        if restart_mask & RESTART_NETWORKING:
            os.system("service networking restart")

        success = True
    except KeyError:
        eventlog.log_d('fail: mac-address')
        reply_fail(REPLY_BAD_REQUEST_400)

    except Exception:
        if try_restore_backup():
            remove_backup()
        raise

    if success:
        remove_backup()
        return reply_done(new_ui_jnode)

    if try_restore_backup():
        remove_backup()
    response_jnode = reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)
    return response_jnode


def _run(query_args):
    """
    :type query_args: UiQueryArgs
    :rtype: dict
    """
    # Dispatch http-method.
    if query_args.http_method == 'GET':
        return _handle_get(query_args)

    elif query_args.http_method == 'PUT':
        try:
            with open(UI_NETWORK_JNODE_FILEPATH, 'r') as settings_file:
                raw_ui_jnode = ujson.load(settings_file)

        except (IOError, ValueError):
            eventlog.log_d('fail: load ui-jnode')
            return reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)
        new_ui_jnode = ujson.loads(query_args.raw_body)
        response_jnode = _handle_put(raw_ui_jnode, new_ui_jnode)

    else:
        response_jnode = reply_fail(REPLY_METHOD_NOT_ALLOWED_405)

    return response_jnode


# noinspection PyBroadException
def _main():
    eventlog.open_log("cgi-network", "debug")
    exit_code = 1
    try:
        query_args = ui_types.UiQueryArgs.deserialize()
        reply_jnode = _run(query_args)

        print ujson.dumps(reply_jnode)
        exit_code = 0
    except Exception:
        eventlog.log_x()
        print ujson.dumps(reply_fail(REPLY_INTERNAL_SERVER_ERROR_500))

    finally:
        eventlog.close_log()

    sys.exit(exit_code)


if __name__ == '__main__':
    _main()
