import ujson
import os

from edge.admin.ui_types import BaseCreator
from edge import pipeline_helpers as oh
from edge import path_helpers as ph
from edge.com import fs

_BASE_SPOOL_PATH = '/var/spool/edge/system'


def _get_cache_dir(ui_pid, edge_gid):
    """
    :type ui_pid: str
    :type edge_gid: str
    :rtype: str
    """
    spool_dir = ph.get_spool_dir(ui_pid)
    return os.path.join(spool_dir, 'bs%s-bs%s-store' % (edge_gid, edge_gid))


def _get_archive_dir(edge_id):
    """
    :type edge_id: str
    :rtype: str
    """
    return "/var/opt/edge/blobstore/%s" % edge_id


def _add_blobstore_channel(blobstore_jnode, channel_args_jnode):
    """
    :type blobstore_jnode: dict
    :type channel_args_jnode: dict
    :rtype dict
    """
    ch_cleanup_jnode = channel_args_jnode['rw']['args']['cleanup']

    new_ch_jnode = {
        "channel": channel_args_jnode['ro']['channel'],
        "cache": {
            "directory": os.path.join(_BASE_SPOOL_PATH, "bs65535-bs65535-store"),
            "limits": {
                "max-du": 50,
                "max-abs-time": 1
            }
        },
        "archive": {
            "directory": _get_archive_dir('65535'),
            "filesize": 4,
            "limits": {
            }
        }
    }

    arc_limits_jnode = new_ch_jnode['archive']['limits']

    if ch_cleanup_jnode['max-du'][0]:
        arc_limits_jnode['max-du'] = ch_cleanup_jnode['max-du'][1]

    if ch_cleanup_jnode['min-df'][0]:
        arc_limits_jnode['min-df'] = ch_cleanup_jnode['min-df'][1]

    if ch_cleanup_jnode['max-abs-time'][0]:
        arc_limits_jnode['max-abs-time'] = ch_cleanup_jnode['max-abs-time'][1]

    if ch_cleanup_jnode['max-rel-time'][0]:
        arc_limits_jnode['max-rel-time'] = ch_cleanup_jnode['max-rel-time'][1]

    blobstore_jnode['storage'].append(new_ch_jnode)

    return blobstore_jnode


def _remove_blobstore_channel(ch_id, blobstore_jnode):
    """
    :type ch_id: str
    :type blobstore_jnode: dict
    :rtype dict
    """
    storages_jnode = blobstore_jnode['storage']
    storages_to_remove = filter(lambda storage_jnode: storage_jnode['channel'] == ch_id, storages_jnode)

    for storage in storages_to_remove:
        storages_jnode.remove(storage)

    return blobstore_jnode


def append_blobstore(channel_args_jnode):
    """
    :type channel_args_jnode: dict
    """
    bs_jnode = fs.load_json(ph.BLOBSTORE_CONF_FILEPATH)
    _add_blobstore_channel(bs_jnode, channel_args_jnode)

    with open(ph.BLOBSTORE_CONF_FILEPATH, 'w') as json_file:
        ujson.dump(bs_jnode, json_file, indent=2)


def remove_blobstore(ch_id):
    """
    :type ch_id: str
    """
    bs_jnode = fs.load_json(ph.BLOBSTORE_CONF_FILEPATH)
    if not bs_jnode:
        raise RuntimeError

    _remove_blobstore_channel(ch_id, bs_jnode)

    with open(ph.BLOBSTORE_CONF_FILEPATH, 'w') as json_file:
        ujson.dump(bs_jnode, json_file, indent=2)


def reconfigure_blobstore(new_jnode):
    """
    :type new_jnode: dict
    """
    blobstore_jnode = fs.load_json(ph.BLOBSTORE_CONF_FILEPATH)
    if not blobstore_jnode:
        raise RuntimeError

    _remove_blobstore_channel(new_jnode['ro']['channel'], blobstore_jnode)
    _add_blobstore_channel(blobstore_jnode, new_jnode)

    with open(ph.BLOBSTORE_CONF_FILEPATH, 'w') as json_file:
        ujson.dump(blobstore_jnode, json_file, indent=2)

    # restart blobstore
    oh.sighup_edge_module(ph.EDGE_TYPE_BLOBSTORE, '65535')


def bs_try_get_creator_ctor(ui_entry_point, used_ui_modules, used_edge_modules, avaliable_ui_modules):
    """
    :type ui_entry_point: UiIdentity
    :type used_ui_modules: set of UiIdentity
    :type used_edge_modules: set of EdgeIdentity
    :type avaliable_ui_modules: set of UiIdentity
    :rtype: None | (str, UiIdentity, dict of UiIdentity: dict) -> BaseCreator
    """
    creator_ctor = None
    if ui_entry_point in [ph.Pipeline.Rtsp.Ui.PIPELINE, ph.Pipeline.Mv.Ui.PIPELINE,
                          ph.Pipeline.Speed.Ui.PIPELINE]:
        creator_ctor = BsCreator.create

    return creator_ctor


class BsCreator(BaseCreator):
    def __init__(self, id_main_host_setupstore):
        """
        :type id_main_host_setupstore: str, dict, str
        """
        BaseCreator.__init__(self)
        self._id_main_setupstore = id_main_host_setupstore

    @staticmethod
    def create(ui_pid, ui_entry_point, ui_identity_to_jnode_map):
        """
        :type ui_pid: str
        :type ui_entry_point: UiIdentity
        :type ui_identity_to_jnode_map: dict of UiIdentity: dict
        :rtype: BsCreator
        """
        if ui_entry_point in [ph.Pipeline.Rtsp.Ui.PIPELINE, ph.Pipeline.Mv.Ui.PIPELINE,
                              ph.Pipeline.Speed.Ui.PIPELINE]:
            id_main_setupstore = BsCreator._serialize(ui_pid, ui_entry_point)
            creator = BsCreator(id_main_setupstore)

        else:
            raise NotImplemented('pipeline not supported')

        return creator

    @staticmethod
    def _serialize(ui_pid, ui_entry_point):
        """
        :type ui_pid: str
        :type ui_entry_point: UiIdentity
        :rtype: str, dict, dict, str
        """
        edge_group = oh.get_edge_group(ui_pid)

        if ui_entry_point.ptype == ph.PTYPE_SPEED:
            edge_identity = ph.Pipeline.Speed.Edge.BLOBSTORE_0
        
        elif ui_entry_point.ptype == ph.PTYPE_LPR_MV:
            edge_identity = ph.Pipeline.Mv.Edge.BLOBSTORE_0

        elif ui_entry_point.ptype == ph.PTYPE_LPR_RTSP:
            edge_identity = ph.Pipeline.Rtsp.Edge.BLOBSTORE_0
            
        else:
            raise NotImplemented

        edge_gid = edge_identity.gid
        edge_id = oh.get_edge_id(edge_group, edge_identity)
        spool_dir = ph.get_spool_dir(ui_pid)

        # create main.conf
        main_conf_jnode = {
            "monitoring": {
                "stat-path": os.path.join(spool_dir, 'bs%s' % edge_gid),
                "stat-tout": 60
            },
            "storage": [
                {
                    "cache": {
                        "directory": _get_cache_dir(ui_pid, edge_gid),
                        "limits": {
                            "max-abs-time": 1,
                            "max-du": 512
                        }
                    },
                    "archive": {
                        "directory": _get_archive_dir(edge_id),
                        "filesize": 4,
                        "limits": {
                            "max-rel-time": 10,
                            "max-du": 10000
                        }
                    },
                    "channel": "0"
                }
            ]
        }

        # create setupstore.conf
        edge_id = oh.get_edge_id(edge_group, edge_identity)
        setupstore_conf_text = ph.get_setupstore_conf_text(edge_id)

        return edge_id, main_conf_jnode, setupstore_conf_text

    def serialize(self):
        edge_id, main_conf_jnode, setupstore_conf_text = self._id_main_setupstore

        edge_jnode_dir = ph.get_edge_jnode_dirpath(ph.EDGE_TYPE_BLOBSTORE, edge_id)
        if not os.path.exists(edge_jnode_dir):
            oh.silent_mkdir_r(edge_jnode_dir)

        main_conf_abspath = os.path.join(edge_jnode_dir, 'main.conf')
        with open(main_conf_abspath, 'w') as json_file:
            ujson.dump(main_conf_jnode, json_file, indent=2)

        # create setupstore.conf
        setupstore_conf_filepath = os.path.join(edge_jnode_dir, 'setupstore.conf')
        with open(setupstore_conf_filepath, 'w') as text_file:
            text_file.write(setupstore_conf_text)


class BsDeleter:
    def __init__(self, edge_bs_dirpath, cache_dir, archive_dir):
        """
        :type edge_bs_dirpath: str
        :type cache_dir: str
        :type archive_dir: str
        """
        self._cache_dir = cache_dir
        self._archive_dir = archive_dir
        self._edge_bs_dirpath = edge_bs_dirpath

    @staticmethod
    def create(ui_pid, ui_entry_point):
        """
        :type ui_pid: str
        :type ui_entry_point: UiIdentity
        :rtype: BsDeleter
        """
        edge_group = oh.get_edge_group(ui_pid)

        if ui_entry_point in [ph.Pipeline.Mv.Ui.PIPELINE, ph.Pipeline.Speed.Ui.PIPELINE,
                              ph.Pipeline.Rtsp.Ui.PIPELINE]:

            edge_id = oh.get_edge_id(edge_group, ph.Pipeline.Mv.Edge.BLOBSTORE_0)  # todo: workaround
            edge_gid = ph.Pipeline.Mv.Edge.BLOBSTORE_0.gid
            conf_dir = ph.get_edge_jnode_dirpath(ph.EDGE_TYPE_BLOBSTORE, edge_id)
            cache_dir = _get_cache_dir(ui_pid, edge_gid)
            archive_dir = _get_archive_dir(edge_id)
            deleter = BsDeleter(conf_dir, cache_dir, archive_dir)
            return deleter

        else:
            raise NotImplemented

    def run(self):
        oh.silent_rm_r(self._edge_bs_dirpath)
        oh.silent_rm_r(self._archive_dir)
        oh.silent_rm_r(self._cache_dir)
