from twisted.internet import reactor
from twisted.internet.protocol import ReconnectingClientFactory, connectionDone
from twisted.protocols.basic import LineOnlyReceiver

from edge.path_helpers import *
from os.path import join as join_path
import _file_lock

import ujson


_QUERY_DONE = 1
_QUERY_INVALID_STATE = 2
_QUERY_INVALID_ARG = 3


def _get_reactor():
    """
    :rtype: IReactorCore
    """
    return reactor


# noinspection PyUnusedLocal
def _get_ra_uds_endpoint(ui_pid):
    """
    :type ui_pid: str
    :rtype: str
    """

    edge_identity = Pipeline.Speed.Edge.RADAR_0
    edge_gid = edge_identity.gid
    spool_dir = get_spool_dir(ui_pid)
    return join_path(spool_dir, "ra%s-ra%s-uds" % (edge_gid, edge_gid))


# noinspection PyMethodMayBeStatic
class IRadar:
    def __init__(self):
        pass

    def query_set_alignment(self, mode, speed, query_arg):
        """
        :type mode: int
        :type speed: int
        :type query_arg: object | None
        """
        raise NotImplemented

    def query_set_location(self, unit_f, unit_l, unit_t,
                           radar_fl, radar_fu, radar_lu,
                           camera_fl, camera_fu, camera_lu,
                           matrix,
                           sensor_wh_mm, sensor_wh_px, focal_dist,
                           query_arg):
        """
        :type unit_f: float
        :type unit_l: float
        :type unit_t: float
        :type radar_fl: float
        :type radar_fu: float
        :type radar_lu: float
        :type camera_fl: float
        :type camera_fu: float
        :type camera_lu: float
        :type matrix: list
        :type sensor_wh_mm: (float, float)
        :type sensor_wh_px: (int, int)
        :type focal_dist: int
        :type query_arg: object | None
        """
        raise NotImplemented


class _UdsTransport(LineOnlyReceiver):
    _UTF8_ENCODING = 'utf-8'
    delimiter = b'\0'

    def __init__(self):
        pass

    def connectionMade(self):
        raise NotImplementedError

    # Called when an established connection is lost.
    def connectionLost(self, reason=connectionDone):
        raise NotImplementedError

    def lineReceived(self, raw_response):
        unicode_reply = raw_response.decode(_UdsTransport._UTF8_ENCODING)
        eventlog.log_d("raw-reply: %s" % unicode_reply)

        try:
            reply_jnode = ujson.loads(unicode_reply)
            self.rx_reply(reply_jnode)
        except ValueError:
            eventlog.log_x()

    def rx_reply(self, reply_jnode):
        raise NotImplementedError

    def tx_query(self, query_jnode):
        """
        :type query_jnode: dict
        """
        if not isinstance(query_jnode, dict):
            eventlog.log_d("fail: invalid query")
            raise ValueError

        unicode_query = ujson.dumps(query_jnode)
        eventlog.log_d("raw-query: %s" % unicode_query)

        utf8_query = unicode_query.encode(_UdsTransport._UTF8_ENCODING)
        self.sendLine(utf8_query)
        return _QUERY_DONE


class _Radar(_UdsTransport, IRadar):
    _QUERY_SET_LOCATION = 1,
    _QUERY_SET_ALIGNMENT = 2

    def __init__(self, radar_handler):
        _UdsTransport.__init__(self)
        IRadar.__init__(self)

        if not isinstance(radar_handler, IRadarHandler):
            eventlog.log_d("fail: invalid value of radar-handler")
            raise ValueError

        self._radar_handler = radar_handler
        self._query_action = None
        self._query_arg = None

    def connectionMade(self):
        eventlog.log_d("done: open ra-socket")
        self._query_arg = None
        self._radar_handler.connection_made(self)

    def connectionLost(self, reason=connectionDone):
        eventlog.log_d("done: close ra-socket % s" % reason)

        query_arg = self._query_arg
        self._query_arg = None

        self._radar_handler.connection_lost(query_arg)

    def rx_reply(self, reply_jnode):
        """
        :type reply_jnode: dict
        """
        if self._query_action == _Radar._QUERY_SET_LOCATION:
            status = reply_jnode['status']

            query_arg = self._query_arg
            self._query_arg = None

            self._radar_handler.handle_set_location(status, query_arg)

        elif self._query_action == _Radar._QUERY_SET_ALIGNMENT:
            status = reply_jnode['status']

            query_arg = self._query_arg
            self._query_arg = None

            self._radar_handler.handle_set_alignment(status, query_arg)

        else:
            eventlog.log_d("fail: invalid query-action")

    def query_set_alignment(self, mode, speed, query_arg):
        """
        :type mode: str
        :type speed: int
        :type query_arg: object
        """
        if self._query_arg:
            eventlog.log_d("fail: another query in progress")
            return _QUERY_INVALID_STATE

        if not query_arg:
            eventlog.log_d("fail: invalid value of query-arg")
            return _QUERY_INVALID_ARG

        self._query_action = _Radar._QUERY_SET_ALIGNMENT
        self._query_arg = query_arg

        query_jnode = {
            'method': 'set',
            "action": "alignment",
            "args": {
                "sensor": {
                    "mode": mode,
                    "mode-alignment": {
                        "speed-filter": speed
                    }
                }
            }
        }

        self.tx_query(query_jnode)
        return _QUERY_DONE

    def query_set_location(self, unit_f, unit_l, unit_u,
                           radar_fl, radar_fu, radar_lu,
                           camera_fl, camera_fu, camera_lu,
                           image_mat,
                           sensor_wh_mm, sensor_wh_px, focal_dist,
                           query_arg):
        """
        :type unit_f: float
        :type unit_l: float
        :type unit_u: float
        :type radar_fl: float
        :type radar_fu: float
        :type radar_lu: float
        :type camera_fl: float
        :type camera_fu: float
        :type camera_lu: float
        :type image_mat: list
        :type sensor_wh_mm: (float, float)
        :type sensor_wh_px: (int, int)
        :type focal_dist: int
        :type query_arg: object
        """
        if self._query_arg:
            eventlog.log_d("fail: another query in progress")
            return _QUERY_INVALID_STATE

        if not query_arg:
            eventlog.log_d("fail: invalid value of query-arg")
            return _QUERY_INVALID_ARG

        self._query_action = _Radar._QUERY_SET_LOCATION
        self._query_arg = query_arg

        query_jnode = {
            'method': 'set',
            'action': 'location',
            'args': {
                'sensor': {
                    'position': {
                        'forward': unit_f,
                        'left': unit_l,
                        'up': unit_u
                    },
                    "rotation": {
                        "elevation": radar_fu,
                        "azimuth": radar_fl,
                        "pitch": radar_lu
                    }
                },
                'camera': {
                    'position': {
                        'forward': unit_f,
                        'left': unit_l,
                        'up': unit_u
                    },
                    'rotation': {
                        'elevation': camera_fu,
                        'azimuth': camera_fl,
                        'pitch': camera_lu
                    },
                    'frustum': {
                        'sensor-wh': sensor_wh_mm,
                        'focal-dist': focal_dist,
                        'forward': [5, 100]
                    },
                    'image-wh': sensor_wh_px
                },
                'image-mapper': {
                    'rhc-to-image-map': image_mat
                }
            }
        }

        self.tx_query(query_jnode)
        return _QUERY_DONE


class _RadarClientFactory(ReconnectingClientFactory):
    initialDelay = 1
    maxDelay = 15
    factor = 1.6180339887498948

    def __init__(self, radar_handler):
        self._radar_action_handler = radar_handler

    def buildProtocol(self, addr):
        self.resetDelay()
        protocol = _Radar(self._radar_action_handler)
        protocol.factory = self
        return protocol

    def clientConnectionFailed(self, connector, reason):
        ReconnectingClientFactory.clientConnectionFailed(self, connector, reason)

    def clientConnectionLost(self, connector, unused_reason):
        ReconnectingClientFactory.clientConnectionLost(self, connector, unused_reason)


# noinspection PyMethodMayBeStatic
class IRadarHandler:
    def __init__(self):
        pass

    def connection_made(self, radar):
        """
        :type radar: IRadar
        """
        raise NotImplemented

    def connection_lost(self, query_arg):
        """
        :type query_arg: object | None
        """
        raise NotImplemented

    def handle_set_location(self, status, query_arg):
        """
        :type status: int
        :type query_arg: object
        """
        raise NotImplemented

    def handle_set_alignment(self, status, query_arg):
        """
        :type status: int
        :type query_arg: object
        """
        raise NotImplemented


class RaLocationEditor(IRadarHandler):
    def __init__(self, ra_lock_dirpath,
                 unit_f, unit_l, unit_u,
                 radar_fl, radar_fu, radar_lu,
                 camera_fl, camera_fu, camera_lu,
                 sensor_wh_mm, sensor_wh_px, focal_dist,
                 image_mat):
        """
        :param unit_f: values in range [-10.0; 10.0], m
        :type unit_f: float

        :param unit_l: values in range [-10.0; 10.0], m
        :type unit_l: float

        :param unit_u: values in range [0; 10.0], m
        :type unit_u: float

        :param radar_fl: values in range [-15.0; 15.0], degrees
        :type radar_fl: float

        :param radar_fu: values in range [0; 9.0], degrees
        :type radar_fu: float

        :param radar_lu: values in range [-5.0; 5.0], degrees
        :type radar_lu: float

        :param camera_fl: values in range [-15.0; 15.0], degrees
        :type camera_fl: float

        :param camera_fu: values in range [0; 9.0], degrees
        :type camera_fu: float

        :param camera_lu: values in range [-5.0; 5.0], degrees
        :type camera_lu: float

        :type image_mat: list

        :param sensor_wh_mm: size of camera sensor in mm
        :type sensor_wh_mm: (float, float)

        :param sensor_wh_px: size of camera sensor in px
        :type sensor_wh_px: (int, int)

        :param focal_dist: focal distance of camera, mm
        :type focal_dist: int
        :type ra_lock_dirpath: str
        """
        IRadarHandler.__init__(self)

        self._radar_lock_dirpath = ra_lock_dirpath

        self._unit_f = unit_f
        self._unit_l = unit_l
        self._unit_u = unit_u

        self._radar_fl = radar_fl  # forward-to-left angle (azimuth)
        self._radar_fu = radar_fu  # forward-to-up angle (elevation)
        self._radar_lu = radar_lu  # left-to-up angle (roll)

        self._camera_fl = camera_fl  # forward-to-left angle (azimuth)
        self._camera_fu = camera_fu  # forward-to-up angle (elevation)
        self._camera_lu = camera_lu  # left-to-up angle (roll)

        self._image_mat = image_mat

        self._sensor_wh_mm = sensor_wh_mm
        self._sensor_wh_px = sensor_wh_px
        self._focal_dist = focal_dist

        self._uds_reply_stat = 2

    def connection_made(self, radar):
        """
        :type radar: IRadar
        """
        radar.query_set_location(
            self._unit_f, self._unit_l, self._unit_u,
            self._radar_fl, self._radar_fu, self._radar_lu,
            self._camera_fl, self._camera_fu, self._camera_lu,
            self._image_mat,
            self._sensor_wh_mm, self._sensor_wh_px, self._focal_dist,
            "str")
        eventlog.log_d('')

    def connection_lost(self, query_arg):
        """
        :type query_arg: object | None
        """
        eventlog.log_d('')

    def handle_set_location(self, status, query_arg):
        """
        :type status: int
        :type query_arg: object
        """
        eventlog.log_d("status: %s" % status)
        self._uds_reply_stat = status
        twisted_reactor = _get_reactor()
        twisted_reactor.stop()

    def handle_set_alignment(self, status, query_arg):
        """
        :type status: int
        :type query_arg: object | None
        """
        eventlog.log_d("status: %s" % status)
        self._uds_reply_stat = status
        twisted_reactor = _get_reactor()
        twisted_reactor.stop()

    @staticmethod
    def deserialize(ui_location_jnode, ui_videosource_jnode, base_lock_dirpath):
        """
        :type ui_location_jnode: dict
        :type ui_videosource_jnode: dict
        :type base_lock_dirpath: str
        :rtype: RaLocationEditor
        """

        eventlog.log_d("new ui jnode %s " % ui_location_jnode)
        # deserialize sensor-position
        camera_jnode = ui_location_jnode['rw']['args']['camera']
        camera_position_jnode = camera_jnode['position']
        unit_f = camera_position_jnode['forward']
        unit_l = camera_position_jnode['left']
        unit_u = camera_position_jnode['up']

        # deserialize sensor-rotation
        radar_rotation_jnode = ui_location_jnode['rw']['args']['radar']['rotation']
        radar_fl = radar_rotation_jnode['azimuth']
        radar_fu = radar_rotation_jnode['elevation']
        radar_lu = radar_rotation_jnode['roll']

        camera_rotation_jnode = camera_jnode['rotation']
        camera_fl = camera_rotation_jnode['azimuth']
        camera_fu = camera_rotation_jnode['elevation']
        camera_lu = camera_rotation_jnode['roll']

        # deserialize camera params
        sensor_wh_mm = ui_videosource_jnode['ro']['sensor-wh-mm']
        sensor_wh_px = ui_videosource_jnode['ro']['sensor-wh-px']
        focal_dist = ui_videosource_jnode['ro']['focal-dist']

        image_mat = ui_location_jnode['ro'].get('rhcc-to-ndc') or ui_location_jnode['rt']['rhcc-to-ndc']
        radar_lock_dirpath = base_lock_dirpath + 'radar/'
        editor = RaLocationEditor(
            radar_lock_dirpath,
            unit_f, unit_l, unit_u,
            radar_fl, radar_fu, radar_lu,
            camera_fl, camera_fu, camera_lu,
            sensor_wh_mm,
            sensor_wh_px,
            focal_dist,
            image_mat)
        return editor

    def serialize(self, ui_pid):
        """
        :type ui_pid: str
        """
        lock = _file_lock.try_lock(self._radar_lock_dirpath, ui_pid)
        try:
            uds_endpoint = _get_ra_uds_endpoint(ui_pid)
            client_factory = _RadarClientFactory(radar_handler=self)
            twisted_reactor = _get_reactor()
            twisted_reactor.connectUNIX(uds_endpoint, client_factory)

            twisted_reactor.callLater(15, twisted_reactor.stop)
            twisted_reactor.addSystemEventTrigger('before', 'shutdown', eventlog.log_d, 'Shutdown...')

            twisted_reactor.run()

            if self._uds_reply_stat != 1:
                raise RuntimeError("fail: invalid uds-reply-stat (%s)" % self._uds_reply_stat)

        finally:
            _file_lock.unlock(lock)
