import math

from edge import eventlog
from edge_bbox.types import mk_plate_bbox, RectF
from edge.image.image import ImEnv, ImEngine, ImSizeFitter
from edge.image.types import ImFooterVars
from edge.service.tz import ITzService
from edge.service.about import IAboutService
from edge.model.types import TimeSpan

from .types import DetectorArgs, RedLightEvidence, RuleArgs
from ..base.best import check_quad_dist_from_img_borders
from ..types import IDetector
from ..com.scene import RlSceneMask
from ..com.types import VideoArgs
from ...model.types import BaseEvent, Event, Waypoint, RedLightViolationWaypoint
from ...model.models import Track, TrTrackItem
from ...services.vs import IVsService, VsImageInfo, VsError
from ...services.tl import ITlService


def points_far_enought(rect1, rect2):
    """
    :type rect1: RectF
    :type rect2: RectF
    :rtype: bool
    """
    plate_h = rect1.fb - rect1.ft
    max_distance = plate_h * 0.7  # % of plate height

    rect1_c_x, rect1_c_y = rect1.center
    rect2_c_x, rect2_c_y = rect2.center

    if math.fabs(rect1_c_x - rect2_c_x) < max_distance and math.fabs(rect1_c_y - rect2_c_y) < max_distance:
        # points are in same rect
        return False
    else:
        # points are far enought
        return True


class _SceneImageInfo(VsImageInfo):
    def __init__(self, tr_item):
        """
        :type tr_item: TrTrackItem
        """
        VsImageInfo.__init__(self, tr_item.lr_item.ts)
        self.tr_item = tr_item


class _RuleServices:
    def __init__(self, tz_service, vs_service, about_service, tl_service):
        """
        :type tz_service: ITzService
        :type vs_service: IVsService
        :type about_service: IAboutService
        :type tl_service: ITlService
        """
        self.tz = tz_service
        self.vs = vs_service
        self.about = about_service
        self.tl = tl_service


class Rule:
    def __init__(self, services, rule_args, detector_name, init_im_env, postbar_init_im_env,
                 postbar_fini_im_env, video_args):
        """
        :type services: _RuleServices
        :type rule_args: RuleArgs
        :type detector_name: str
        :type init_im_env: ImEnv
        :type postbar_init_im_env: ImEnv
        :type postbar_fini_im_env: ImEnv
        :type video_args: VideoArgs | None
        :type
        """
        self._detector_name = detector_name

        self._services = services
        self._args = rule_args
        self._pre_im_env = init_im_env
        self._post_im_env1 = postbar_init_im_env
        self._post_im_env2 = postbar_fini_im_env

        self._video = video_args
        if video_args:
            self._im_fitter = ImSizeFitter((video_args.max_w, video_args.max_h))

    def check(self, track, base, pre_stopbar, on_stopbar, post_stopbar):
        """
        :type track: Track
        :type base: BaseEvent
        :type pre_stopbar: list[TrTrackItem]
        :type on_stopbar: list[TrTrackItem]
        :type post_stopbar: list[TrTrackItem]
        :rtype: Evidence | None
        """
        tl_result = self._services.tl.check(track.lr_track.min_ts, track.lr_track.max_ts)
        for violation in tl_result:

            # check pre-stopbar conditions
            pre_ts_range = violation.intersect(
                TimeSpan(pre_stopbar[0].ts, pre_stopbar[-1].ts))
            if pre_ts_range is None:
                continue

            pre_tr_items = filter(
                lambda tr_item: tr_item.ts in pre_ts_range, pre_stopbar)
            if not pre_tr_items:
                continue

            # check on-stopbar conditions
            on_ts_range = violation.intersect(
                TimeSpan(on_stopbar[0].ts, on_stopbar[-1].ts))
            if on_ts_range is None:
                continue

            on_tr_items = filter(
                lambda tr_item: tr_item.ts in on_ts_range, on_stopbar)
            if not on_tr_items:
                continue

            # check post-stopbar conditions
            post_ts_range = violation.intersect(
                TimeSpan(post_stopbar[0].ts, post_stopbar[-1].ts))
            if post_ts_range is None:
                continue

            post_tr_items = filter(
                lambda tr_item: tr_item.ts in post_ts_range, post_stopbar)
            if not post_tr_items:
                continue

            # prepare im-environments
            pre_im_infos = map(_SceneImageInfo, pre_tr_items)
            post_im_infos = map(_SceneImageInfo, post_tr_items)

            pre_im_info = self._services.vs.export_image_rgba2(
                self._pre_im_env.scene_rgba, pre_im_infos[:-6:-1])  # type: _SceneImageInfo
            if pre_im_info is None:
                eventlog.log_d('fail: load pre-stopbar image')
                continue

            post_im_info1 = self._services.vs.export_image_rgba2(
                self._post_im_env1.scene_rgba, post_im_infos[0:5])  # type: _SceneImageInfo
            if post_im_info1 is None:
                eventlog.log_d('fail: load post-stopbar image-1')
                continue

            post_im_info2 = self._services.vs.export_image_rgba2(
                self._post_im_env2.scene_rgba, post_im_infos[:-6:-1])  # type: _SceneImageInfo
            if post_im_info2 is None:
                eventlog.log_d('fail: load post-stopbar image-2')
                continue

            # prepare evidence
            evidence = self._mk_evidence(
                base, track, pre_im_info.tr_item,
                post_im_info1.tr_item, post_im_info2.tr_item)

            if evidence:
                # attach videorecord
                if self._video:
                    scaled_w, scaled_h = self._im_fitter.scaledown(self._pre_im_env.scene_rgba.wh)
                    try:
                        ref_ts = post_tr_items[0].lr_item.ts / 1000
                        rl_video = self._services.vs.export_video(
                            ref_ts - self._video.pre_record,
                            ref_ts + self._video.post_record,
                            scaled_w, scaled_h, self._video.quality)
                        evidence.add_videos([rl_video])

                    except VsError:
                        eventlog.log_d('fail: mk video-blobs')

                return evidence

    def _mk_evidence(self, base_event, track, pre_tr_item, post_tr_item1, post_tr_item2):
        """
        :type base_event: BaseEvent
        :type track: Track
        :type pre_tr_item: TrTrackItem
        :type post_tr_item1: TrTrackItem
        :type post_tr_item2: TrTrackItem
        :rtype: RedLightEvidence | None
        """
        plate_text = base_event.get_plate_text()

        def mk_waypoint(services, im_env, tr_item, waypoint_name):
            """
            :type services: _RuleServices
            :type im_env: ImEnv
            :type tr_item: TrTrackItem
            :type waypoint_name: str
            :rtype: Waypoint
            """
            lr_scene_img = track.lr_track.best_image
            plate_rect = RectF(*tr_item.lr_item.plate_quad.get_rel_bbox(lr_scene_img.width, lr_scene_img.height))
            plate_bbox = mk_plate_bbox(plate_rect, *im_env.scene_rgba.wh)
            origin = services.about.get_origin(im_env.scene_rgba.ts)

            footer_vars = ImFooterVars(services.tz, im_env.scene_rgba.ts, im_env.footer_def_vars,
                                       location=origin.location, geolocation=origin.geolocation, plate_text=plate_text,
                                       spot_speed=tr_item.ra_item.speed if tr_item.ra_item else None)

            im_blobs = im_env.engine.mk_blobs(
                im_env.scene_rgba, plate_bbox, footer_vars, track.lr_track, tr_item.lr_item.plate_quad,
                vehicle=False, scene=True, plate=True)

            if len(im_blobs) != 2:
                eventlog.log_d('fail: mk image-blobs')
                raise RuntimeError

            waypoint = RedLightViolationWaypoint(
                waypoint_name, im_env.scene_rgba.ts, origin.location, [tr_item.lr_item.plate_quad],
                *services.tl.get_times(im_env.scene_rgba.ts))

            waypoint.add_images(im_blobs)
            return waypoint

        try:
            pre_waypoint = mk_waypoint(
                self._services, self._pre_im_env, pre_tr_item,
                Waypoint.NAME_PRE_STOPBAR)

            post_waypoint1 = mk_waypoint(
                self._services, self._post_im_env1, post_tr_item1,
                Waypoint.NAME_POST_STOPBAR)

            post_waypoint2 = mk_waypoint(
                self._services, self._post_im_env2, post_tr_item2,
                Waypoint.NAME_POST_STOPBAR)

        except RuntimeError:
            return None

        event_ts = self._post_im_env1.scene_rgba.ts

        evidence = RedLightEvidence(
            self._services.about.get_origin(event_ts),
            event_ts, self._args, self._detector_name)
        evidence.add_waypoints([pre_waypoint, post_waypoint1, post_waypoint2])
        return evidence


def create(*args, **kwargs):
    """
    :rtype: IDetector
    """
    return _Detector(*args, **kwargs)


class _Detector(IDetector):
    def __init__(self, tz_service, vs_service, about_service, tl_service, args):
        """
        :type vs_service: IVsService
        :type tz_service: ITzService
        :type tl_service: ITlService
        :type args: DetectorArgs
        """
        IDetector.__init__(self)

        self._sc_detector = RlSceneMask(args.scene)

        im_engine = ImEngine(args.im_engine)
        prebar_im_env = ImEnv(
            im_engine, args.im_engine.scene.footer.height, args.def_footer_mapping)

        postbar_init_im_env = ImEnv(
            im_engine, args.im_engine.scene.footer.height, args.def_footer_mapping)

        postbar_fini_im_env = ImEnv(
            im_engine, args.im_engine.scene.footer.height, args.def_footer_mapping)

        self._rules = []

        rule_services = _RuleServices(tz_service, vs_service, about_service, tl_service)
        for rule in args.rules:
            rule = Rule(
                rule_services, rule, args.name, prebar_im_env, postbar_init_im_env,
                postbar_fini_im_env, args.video)
            self._rules.append(rule)

        self._identity = args.identity
        self._handler = None

    def set_handler(self, event_handler):
        """
        :type event_handler: (Event) -> None
        """
        if self._handler:
            raise NotImplementedError

        self._handler = event_handler

    def analyze(self, track, base):
        """
        :type track: Track
        :type base: BaseEvent
        """
        if track.lr_track is None:
            return

        def split_track(sc_detector):
            """
            :type sc_detector: RlSceneMask
            :rtype: (list[TrTrackItem], list[TrTrackItem], list[TrTrackItem]) | None
            """
            lr_scene_img = track.lr_track.best_image
            tr_items = filter(
                lambda item: item.lr_item and check_quad_dist_from_img_borders(
                    item.lr_item.plate_quad, lr_scene_img.width, lr_scene_img.height),
                track.tr_track)

            pre_stopbar, on_stopbar, post_stop_bar = [], [], []
            for tr_item in tr_items:
                check_answer = sc_detector.check_quad(tr_item.lr_item.plate_quad)

                if check_answer == sc_detector.STOPBAR_PRE:
                    pre_stopbar.append(tr_item)

                elif check_answer == sc_detector.STOPBAR_ON:
                    if len(pre_stopbar):
                        on_stopbar.append(tr_item)

                    else:
                        # we didn't have pre-stopbar items yet
                        # car direction is wrong or we din't have points before stop bar
                        # print 'err in'
                        return None

                elif check_answer == sc_detector.STOPBAR_POST:
                    if len(pre_stopbar) == 0 or len(on_stopbar) == 0:
                        # we didn't have pre-stop-bar or inside-stop-bar items yet
                        # car direction is wrong or we din't have points before stop bar
                        # print 'err post'
                        return None

                    post_stop_bar.append(tr_item)

            return pre_stopbar, on_stopbar, post_stop_bar

        answer = split_track(self._sc_detector)
        if answer is None:
            # controlling direction of movement is not defined
            return

        elif 0 in map(len, answer):
            # stopbar crossing is not detected
            return

        for rule in self._rules:
            evidence = rule.check(track, base, *answer)
            if evidence:
                self._handler(Event(self._identity, base, evidence))
                break
