import ctypes

from glob import glob
from edge import eventlog
from edge.types import BufferSizeError
from .types import SignServiceArgs


class SignError(RuntimeError):
    pass


class _SignSessionError(SignError):
    pass


class ISignService:
    def __init__(self):
        pass

    def sign_blob(self, blob_buffer, sign_buffer):
        """
        :type blob_buffer: str
        :type sign_buffer: str

        :rtype: int
        :raise SignError:
        """
        raise NotImplementedError

    def hash_blob(self, blob_buffer, hash_buffer):
        """
        :type blob_buffer: str
        :type hash_buffer: str

        :rtype: int
        :raise SignError:
        """
        raise NotImplementedError


def init_service(args):
    """
    :type args: ISignServiceArgs
    :rtype: ISignService
    """
    return _SignService(args)


def fini_service(service):
    """
    :type service: ISignService | _SignService
    """
    service.destroy()


class _Api:
    LOG_N = 0
    LOG_E = 1
    LOG_W = 2
    LOG_I = 3
    LOG_D = 4
    LOG_T = 5

    RC_DONE = 0
    RC_FAIL = -1
    RC_DENY = -2

    _lib = ctypes.CDLL(glob('/opt/edge/lib/libsign*.so')[0])

    # int
    # EsOpen(const char *aCfgFile, uint8_t aLogLevel, void **aSession);
    open = _lib.EsOpen
    open.restype = ctypes.c_int
    open.argtypes = (ctypes.c_char_p, ctypes.c_uint8, ctypes.POINTER(ctypes.c_voidp))

    # int
    # EsSign2(void *aSession, const uint8_t *aBlobData, uint32_t aBlobSize,
    #     uint8_t *aSignData, uint32_t *aSignSize);
    sign2 = _lib.EsSign2
    sign2.restype = ctypes.c_int
    sign2.argtypes = (ctypes.c_voidp, ctypes.POINTER(ctypes.c_uint8), ctypes.c_uint32,
                      ctypes.POINTER(ctypes.c_uint8), ctypes.POINTER(ctypes.c_uint32))

    # int
    # EsHash2 (void *aSession, const uint8_t *aBlobData, uint32_t aBlobSize,
    #     uint8_t *aHashData, uint32_t *aHashSize);
    hash2 = _lib.EsHash2
    hash2.restype = ctypes.c_int
    hash2.argtypes = (ctypes.c_voidp, ctypes.POINTER(ctypes.c_uint8), ctypes.c_uint32,
                      ctypes.POINTER(ctypes.c_uint8), ctypes.POINTER(ctypes.c_uint32))

    # void
    # EsClose (void *aSession);
    close = _lib.EsClose
    close.argtype = ctypes.c_voidp

    def __init__(self):
        pass


def _check_session(fn):
    def proxy_fn(self, *args, **kwargs):
        """
        :type self: _Session
        :raise _SignSessionError:
        """
        if self.closed:
            eventlog.log_e('fail: invalid session')
            raise _SignSessionError

        return fn(self, *args, **kwargs)

    return proxy_fn


class _Session:
    def __init__(self, handle):
        """
        :type handle: ctypes.c_void_p | None
        """
        self._handle = handle
        self.closed = handle is None

    @staticmethod
    def open(conf_filepath, log_level=_Api.LOG_I):
        """
        :type conf_filepath: str
        :type log_level: int
        :rtype: _Session

        :raise VsSessionError:
        """
        handle = ctypes.c_void_p(0)
        rc = _Api.open(conf_filepath, log_level, handle)
        if rc != _Api.RC_DONE:
            eventlog.log_e('fail: EsOpen <{}>'.format(rc))
            raise _SignSessionError

        return _Session(handle=handle)

    @_check_session
    def sign_buffer(self, blob_buffer, sign_buffer):
        """
        :type blob_buffer: str
        :param sign_buffer: ctypes.create_string_buffer
        :type sign_buffer: str

        :rtype: int
        :raise SignError:
        """
        blob_data = ctypes.cast(blob_buffer, ctypes.POINTER(ctypes.c_uint8))
        blob_size = ctypes.c_uint32(len(blob_buffer))

        sign_data = ctypes.cast(sign_buffer, ctypes.POINTER(ctypes.c_uint8))

        init_sign_size = len(sign_buffer)
        sign_size = ctypes.c_uint32(init_sign_size)
        p_sign_size = ctypes.pointer(sign_size)

        rc = _Api.sign2(self._handle, blob_data, blob_size, sign_data, p_sign_size)
        if rc != _Api.RC_DONE:
            eventlog.log_e('fail: EsSign <{}>'.format(rc))
            raise SignError

        fini_sign_size = sign_size.value
        if fini_sign_size > init_sign_size:
            eventlog.log_e('fail: sign-buffer too small (min-size: {})'.format(fini_sign_size))
            raise BufferSizeError(fini_sign_size)

        return fini_sign_size

    @_check_session
    def hash_buffer(self, blob_buffer, hash_buffer):
        """
        :type blob_buffer: str
        :param hash_buffer: ctypes.create_string_buffer
        :type hash_buffer: str

        :rtype: int
        :raise SignError:
        """
        blob_data = ctypes.cast(blob_buffer, ctypes.POINTER(ctypes.c_uint8))
        blob_size = ctypes.c_uint32(len(blob_buffer))

        hash_data = ctypes.cast(hash_buffer, ctypes.POINTER(ctypes.c_uint8))

        init_hash_size = len(hash_buffer)
        hash_size = ctypes.c_uint32(init_hash_size)
        p_hash_size = ctypes.pointer(hash_size)

        rc = _Api.hash2(self._handle, blob_data, blob_size, hash_data, p_hash_size)
        if rc != _Api.RC_DONE:
            eventlog.log_e('fail: EsHash (%s)' % rc)
            raise SignError

        fini_hash_size = hash_size.value
        if fini_hash_size > init_hash_size:
            eventlog.log_e('fail: hash-buffer too small (min-size: {})'.format(fini_hash_size))
            raise BufferSizeError(fini_hash_size)

        return fini_hash_size

    def close(self):
        _Api.close(self._handle)
        self.closed = True


def _reopen_session(fn):
    def proxy_fn(self, *args, **kwargs):
        """
        :type self: _SignService
        """
        try_count = 0
        while try_count <= 3:
            try_count += 1
            try:
                if self._session.closed:
                    self._session = _Session.open(*self._session_args)

                return fn(self, *args, **kwargs)

            except _SignSessionError:
                self._session.close()

        raise SignError

    return proxy_fn


_LOG_LEVEL_NAME_TO_CODE_MAP = {
    'none': _Api.LOG_N,
    'error': _Api.LOG_E,
    'warn': _Api.LOG_W,
    'info': _Api.LOG_I,
    'debug': _Api.LOG_D,
    'trace': _Api.LOG_T
}


class _SignService(ISignService):
    def __init__(self, args):
        """
        :type args: SignServiceArgs
        """
        ISignService.__init__(self)
        self._session_args = args.conf_path, _LOG_LEVEL_NAME_TO_CODE_MAP.get(args.log_level, 'info')
        self._session = _Session(handle=None)

    def destroy(self):
        self._session.close()

    @_reopen_session
    def sign_blob(self, blob_buffer, sign_buffer):
        sign_size = self._session.sign_buffer(blob_buffer, sign_buffer)
        return sign_size

    @_reopen_session
    def hash_blob(self, blob_buffer, hash_buffer):
        hash_size = self._session.hash_buffer(blob_buffer, hash_buffer)
        return hash_size
