import os
import ctypes

from glob import glob
from edge import eventlog
from edge.image.image import ImRgba
from edge.image.types import Blob, FileBlob, JsonBlobMeta

from .types import VsServiceArgs

from tempfile import NamedTemporaryFile


class VsImageInfo:
    def __init__(self, ts):
        """
        :type ts: long
        """
        self._ts = ts

    @property
    def ts(self):
        """
        :return: value in usec
        :rtype: long
        """
        return self._ts


class VsBufferTooSmall(RuntimeError):
    def __init__(self, min_w, min_h):
        """
        :type min_w:
        :type min_h:
        """
        self.min_w = min_w
        self.min_h = min_h


class VsError(RuntimeError):
    pass


class VsSessionError(VsError):
    pass


class IVsService:
    def __init__(self):
        pass

    def export_image_rgba(self, im_rgba, im_infos):
        """
        :type im_rgba: ImRgba
        :type im_infos: list[VsImageInfo]

        :rtype: VsImageInfo | None

        :raise VsBufferTooSmall:
        :raise VsError:
        """
        raise NotImplementedError

    def export_image_rgba2(self, im_rgba, im_infos):
        """
        :type im_rgba: ImRgba
        :type im_infos: list[VsImageInfo]

        :rtype: VsImageInfo | None

        :raise VsError:
        """
        while True:
            try:
                return self.export_image_rgba(im_rgba, im_infos)
            except VsBufferTooSmall as x:
                min_w, min_h = x.min_w, x.min_h

            im_rgba.reset(im_rgba.ts, min_w, min_h, im_rgba.dh)

    def export_video(self, min_ts, max_ts, max_width, max_height, rec_quality):
        """
        :param min_ts: ms since 01.01.1970 00:00 UTC
        :type min_ts: long

        :param max_ts: ms since 01.01.1970 00:00 UTC
        :type max_ts: long

        :param max_width: valid values in range [16; 32752], multiple 16.
        :type max_width: int

        :param max_height: valid values in range [16; 32752], multiple 16.
        :type max_height: int

        :param rec_quality: Constant Bit Rate (CBR), Kbit/s. Valid values in range [100; 102400].
        :type rec_quality: int

        :rtype: Blob
        :raise VsError:
        """
        raise NotImplementedError


def init_service(args):
    """
    :type args: VsServiceArgs
    :rtype: IVsService
    """
    return _VsService(args)


def fini_service(service):
    """
    :type service: IVsService | _VsService
    """
    service.destroy()


class _Api:
    RC_DONE = 0
    RC_FAIL = -1
    RC_DENY = -2
    RC_CONN = -3

    LOG_N = 0
    LOG_E = 1
    LOG_W = 2
    LOG_I = 3
    LOG_D = 4
    LOG_T = 5

    IMAGE_GRAY = 1
    IMAGE_RGBA = 2
    IMAGE_JPEG = 3

    VIDEO_H264_AVI = 1
    VIDEO_H264_MKV = 2
    VIDEO_H264_MP4 = 3

    _lib = ctypes.CDLL(glob('/opt/edge/lib/libblobstore*.so')[0])

    # int
    # BsvOpen(uint16_t aStore, uint8_t aLogLevel, void **aSession);

    open = _lib.BsvOpen
    open.argtypes = (ctypes.c_uint16, ctypes.c_uint8, ctypes.POINTER(ctypes.c_voidp))
    open.restype = ctypes.c_int

    # int
    # BsvLoadImage(void *aSession, const char *aChannel, uint64_t anImTs, uint8_t anImFormat, uint8_t *anImData,
    #     uint32_t *anImSize,uint16_t *anImWidth, uint16_t *anImHeight);

    load = _lib.BsvLoadImage
    load.argtypes = (ctypes.c_voidp, ctypes.c_char_p, ctypes.c_uint64, ctypes.c_uint8, ctypes.POINTER(ctypes.c_uint8),
                     ctypes.POINTER(ctypes.c_uint32), ctypes.POINTER(ctypes.c_uint16), ctypes.POINTER(ctypes.c_uint16))

    load.restype = ctypes.c_int

    # int
    # BsvLoadRecord(void *aSession, const char *aChannel, uint64_t *aRecMinTs, uint64_t *aRecMaxTs,
    #     uint8_t aRecFormat, uint32_t aRecQuality, uint16_t aRecSkipRate,
    #     uint16_t aRecMaxWidth, uint16_t aRecMaxHeight, const char *aRecPath);

    save = _lib.BsvLoadRecord
    save.argtypes = (ctypes.c_voidp, ctypes.c_char_p, ctypes.POINTER(ctypes.c_uint64), ctypes.POINTER(ctypes.c_uint64),
                     ctypes.c_uint8, ctypes.c_uint32, ctypes.c_uint16,
                     ctypes.c_uint16, ctypes.c_uint16, ctypes.c_char_p)

    save.restype = ctypes.c_int

    # void
    # BsvClose(void *aSession);
    close = _lib.BsvClose
    close.argtype = ctypes.c_voidp

    def __init__(self):
        pass


def _check_session(fn):
    def proxy_fn(self, *args, **kwargs):
        """
        :type self: _Session
        """
        if self.closed:
            eventlog.log_e('fail: invalid session')
            raise VsSessionError

        return fn(self, *args, **kwargs)

    return proxy_fn


class _Session:
    def __init__(self, handle, chan):
        """
        :type handle: ctypes.c_void_p | None
        :type chan: str
        """
        self._chan = chan
        self._handle = handle
        self.closed = handle is None

    @staticmethod
    def open(store, chan, log_level=_Api.LOG_I):
        """
        :type store: int
        :type chan: str
        :type log_level: int
        :rtype: _Session

        :raise VsSessionError:
        """
        handle = ctypes.c_void_p(0)
        rc = _Api.open(store, log_level, handle)

        if rc != _Api.RC_DONE:
            eventlog.log_e('fail: BsvOpen (%s)' % rc)
            raise VsSessionError

        return _Session(handle=handle, chan=chan)

    @_check_session
    def export_image_rgba(self, im_ts, im_buffer):
        """
        :param im_ts: value in msec
        :type im_ts: long

        :type im_buffer: str
        :rtype: (int, int)
        
        :raise VsError:
        :raise VsSessionError:
        """
        im_data = ctypes.cast(im_buffer, ctypes.POINTER(ctypes.c_uint8))
        im_size = ctypes.c_uint32(ctypes.sizeof(im_buffer))
        im_w, im_h = ctypes.c_uint16(), ctypes.c_uint16()

        rc = _Api.load(
            self._handle, self._chan, ctypes.c_uint64(im_ts), ctypes.c_uint8(_Api.IMAGE_RGBA),
            im_data, im_size, im_w, im_h)

        if im_size.value > ctypes.sizeof(im_buffer):
            eventlog.log_e('fail: buffer too small')
            return im_w.value, im_h.value

        if rc == _Api.RC_CONN:
            eventlog.log_e('fail: BsvLoadImage (kBS_CONN)')
            raise VsSessionError

        elif rc != _Api.RC_DONE:
            eventlog.log_e('fail: BsvLoadImage (%s)' % rc)
            raise VsError

        return im_w.value, im_h.value

    @_check_session
    def export_video(self, min_ts, max_ts, max_width, max_height, rec_quality):
        """
        :param min_ts: ms since 01.01.1970 00:00 UTC
        :type min_ts: long

        :param max_ts: ms since 01.01.1970 00:00 UTC
        :type max_ts: long

        :param max_width: valid values in range [16; 32752], multiple 16.
        :type max_width: int

        :param max_height: valid values in range [16; 32752], multiple 16.
        :type max_height: int

        :param rec_quality: Constant Bit Rate (CBR), Kbit/s. Valid values in range [100; 102400].
        :type rec_quality: int

        :rtype: Blob

        :raise VsError:
        :raise VsSessionError:
        """
        min_time_buf, max_time_buf = ctypes.c_uint64(min_ts), ctypes.c_uint64(max_ts)
        video_file = NamedTemporaryFile()

        rc = _Api.save(
            self._handle, self._chan, min_time_buf, max_time_buf,
            ctypes.c_uint8(_Api.VIDEO_H264_MP4), ctypes.c_uint32(rec_quality), ctypes.c_uint16(1),
            ctypes.c_uint16(max_width), ctypes.c_uint16(max_height),
            ctypes.cast(video_file.name, ctypes.POINTER(ctypes.c_char)))

        if rc == _Api.RC_CONN:
            eventlog.log_e('fail: BsvLoadRecord (kBS_CONN)')
            raise VsSessionError

        elif rc != _Api.RC_DONE:
            eventlog.log_e('fail: BsvLoadRecord (%s)' % rc)
            raise VsError

        blob_meta = JsonBlobMeta({
            "size": [max_width, max_height, os.path.getsize(video_file.name)]
        })
        return FileBlob(min_time_buf.value * 1000, max_time_buf.value * 1000, video_file, meta=blob_meta)

    def close(self):
        _Api.close(self._handle)
        self.closed = True


def _reopen_session(fn):
    def proxy_fn(self, *args, **kwargs):
        """
        :type self: _VsService
        """
        try_count = 0
        while try_count <= 3:
            try_count += 1
            try:
                if self._session.closed:
                    self._session = _Session.open(*self._session_args)

                return fn(self, *args, **kwargs)

            except VsSessionError:
                self._session.close()

        raise VsError

    return proxy_fn


_LOG_LEVEL_NAME_TO_CODE_MAP = {
    'none': _Api.LOG_N,
    'error': _Api.LOG_E,
    'warn': _Api.LOG_W,
    'info': _Api.LOG_I,
    'debug': _Api.LOG_D,
    'trace': _Api.LOG_T
}


class _VsService(IVsService):
    def __init__(self, args):
        """
        :type args: VsServiceArgs
        """
        IVsService.__init__(self)
        self._session_args = args.store, args.chan, _LOG_LEVEL_NAME_TO_CODE_MAP.get(args.log_level, 'info')
        self._session = _Session(handle=None, chan=args.chan)

    def destroy(self):
        self._session.close()

    @_reopen_session
    def export_image_rgba(self, im_rgba, im_infos):
        """
        :type im_rgba: ImRgba
        :type im_infos: list[VsImageInfo]

        :rtype: VsImageInfo | None

        :raise VsBufferTooSmall:
        :raise VsError:
        """
        for im_info in im_infos:
            try:
                image_w, image_h = self._session.export_image_rgba(im_info.ts / 1000, im_rgba.buffer)

            except VsSessionError:
                raise

            except VsError:
                eventlog.log_i('fail: export image-ts=<%s> (trying the next one)' % im_info.ts)
                continue

            if image_w > im_rgba.w or image_h > im_rgba.h:
                raise VsBufferTooSmall(image_w, image_h)

            elif image_w > 0 and image_h > 0:
                im_rgba.reset(im_info.ts, image_w, image_h, im_rgba.dh)
                return im_info

            else:
                eventlog.log_e('fail: invalid image-wh=<%sx%s>' % (image_w, image_h))
                raise VsError

    @_reopen_session
    def export_video(self, min_ts, max_ts, max_width, max_height, rec_quality):
        return self._session.export_video(min_ts, max_ts, max_width, max_height, rec_quality)
